const chalk = require('chalk');

let defaultOpts = {
    scope: {
        mainStyle: 'grey.dim',
        subStyle: 'grey.dim',
        mainEnabled: true,
        subEnabled: true
    },
    fail: {
        text: 'Failure',
        icon: {
            text: '✕',
            style: 'redBright.bold'
        },
        style: 'red.underline',
        testDescStyle: 'redBright.italic',
        detailStyle: 'grey',
        detailValueStyle: 'yellow'
    },
    success: {
        text: 'Success',
        icon: {
            text: '✓',
            style: 'greenBright.bold'
        },
        style: 'green.underline',
        testDescStyle: 'greenBright.italic',
        detailStyle: 'grey',
        detailValueStyle: 'yellow'
    }
}

function generateStyle(styleString) {
    try {
        let style = chalk;
        styleString.split('.').forEach(s => {
            if(s.match(/hex\(#.{1,}\)/g)){
                style = style.hex(s.split('(')[1].split('').reverse().join('').substr(1).split('').reverse().join(''));
            }else{
                style = style[s];
            }
        });
        return style;
    } catch (e) {
        throw new Error(`'${styleString} is not a valid style literal. Note: this can be caused by atempting to use rgb/hex functions.`);
    }
}

function optOrDef(opt, def) {
    if (opt == undefined) return def;
    else return opt;
}

Object.defineProperty(global, '__file', {
    get: function () {
        var orig = Error.prepareStackTrace;
        Error.prepareStackTrace = function (_, stack) {
            return stack;
        };
        var err = new Error();
        Error.captureStackTrace(err, arguments.callee);
        var stack = err.stack;
        Error.prepareStackTrace = orig;

        let currentfile = err.stack.shift();
        let callerfile;
        while (err.stack.length) {
            callerfile = err.stack.shift();

            if (currentfile !== callerfile) break;
        }
        return callerfile;
    },
});



class subTest {
    constructor(description, name, detail, parent) {
        this.name = name;
        this.description = description;
        this.parent = parent;
        this.detail = detail;
        this.successValueStyle = generateStyle(parent.opts.success.detailValueStyle || defaultOpts.success.detailValueStyle);
        this.failValueStyle = generateStyle(parent.opts.success.detailValueStyle || defaultOpts.fail.detailValueStyle);
    }

    equals() {
        let args = Array.prototype.slice.call(arguments, 0);
        if (args.every(v => v === args[0])) {
            this.parent._success(this, __file, `\r`);
        } else {
            let detail;
            if (!this.parent.detail && this.detail || this.parent.detail && this.detail) {
                detail =
                    `\n\rValues were not equivelent:
                \r${args.map(v => this.failValueStyle(require('util').inspect(v))).join('\nand\n')}`;
            } else {
                detail = `Values were not equivelent.`;
            }
            this.parent._fail(this, __file, detail);
        }
    }

    notEqual() {
        let args = Array.prototype.slice.call(arguments, 0);
        if (args.every(v => v === args[0])) {
            let detail;
            if (!this.parent.detail && this.detail || this.parent.detail && this.detail) {
                detail =
                    `\n\rValues were equivelent:
                \r${args.map(v => this.failValueStyle(require('util').inspect(v)) + `: ${typeof v}`).join('\nand\n')}`;
            } else {
                detail = `Values were equivelent.`;
            }
            this.parent._fail(this, __file, detail);
        } else {
            this.parent._success(this, __file, ``);
        }
    }

    sameType() {
        let args = Array.prototype.slice.call(arguments, 0);
        if (args.every(v => typeof v === typeof args[0])) {
            this.parent._success(this, __file, ``);
        } else {
            let detail;
            if (!this.parent.detail && this.detail || this.parent.detail && this.detail) {
                detail =
                    `\n\rValues were not the same type:
                \r${args.map(v => this.failValueStyle(require('util').inspect(v))).join('\nand\n')}`;
            } else {
                detail = `Values were not the same type.`;
            }
            this.parent._fail(this, __file, detail);
        }
    }

    differentType() {
        let args = Array.prototype.slice.call(arguments, 0);
        if (args.every(v => typeof v === typeof args[0])) {
            let detail;
            if (!this.parent.detail && this.detail || this.parent.detail && this.detail) {
                detail =
                    `\n\rValues were the same type:
                \r${args.map(v => this.failValueStyle(require('util').inspect(v)) + `: ${typeof v}`).join('\nand\n')}`;
            } else {
                detail = `Values were the same type.`;
            }
            this.parent._fail(this, __file, detail);
        } else {
            this.parent._success(this, __file, ``);
        }
    }
}

class Test {
    constructor(name, detail, opts = { scope: {}, fail: { icon: {} }, success: { icon: {} } }) {
        this.name = name;
        this.detail = detail;
        this.opts = opts;
    }

    test(detail, name, description) {
        return new subTest(description, name, detail, this);
    }

    _success(child, file, detail) {
        //Scope Generation
        let mainScopeColor = generateStyle(this.opts.scope.mainStyle || defaultOpts.scope.mainStyle);
        let subScopeColor = generateStyle(this.opts.scope.subStyle || defaultOpts.scope.subStyle);
        let scopes = [];
        if (optOrDef(this.opts.scope.mainEnabled, defaultOpts.scope.mainEnabled)) scopes = scopes.concat([
            mainScopeColor(
                `[${
                this.name
                ||
                require('path').basename(file.getFileName())
                }]`
            )
        ]);
        if (optOrDef(this.opts.scope.subEnabled, defaultOpts.scope.subEnabled) && child.name) scopes = scopes.concat([
            subScopeColor(
                `[${child.name}]`
            )
        ]);
        //End of Scope Generation

        //Icon Generation
        let iconStyle = generateStyle(this.opts.success.icon.style || defaultOpts.success.icon.style);
        let icon;
        if (this.opts.success.icon.text) icon = this.opts.success.icon.text + ' ';
        else icon = defaultOpts.success.icon.text + ' ';

        if (icon == ' ') icon = '';

        icon = iconStyle(icon);
        //End of Icon Generation

        //Type Name Generation
        let nameStyle = generateStyle(this.opts.success.style || defaultOpts.success.style);
        let name = this.opts.success.text || defaultOpts.success.text;

        name = nameStyle(name);
        //End of Name Generation

        //Description Generation
        let descStyle = generateStyle(this.opts.success.testDescStyle || defaultOpts.success.testDescStyle);
        let description = child.description || `Line ${file.getLineNumber()}`;

        description = descStyle(description);
        //End of Description Generation

        //Detail Generation
        let detailStyle = generateStyle(this.opts.success.detailStyle || defaultOpts.success.detailStyle);
        let detailText = detailStyle(detail);
        //End of Detail Generation

        console.log(
            `${scopes.join(' ')} ${icon}${name} ${description} ${detailText}`
        );
    }

    _fail(child, file, detail) {
        //Scope Generation
        let mainScopeColor = generateStyle(this.opts.scope.mainStyle || defaultOpts.scope.mainStyle);
        let subScopeColor = generateStyle(this.opts.scope.subStyle || defaultOpts.scope.subStyle);
        let scopes = [];
        if (optOrDef(this.opts.scope.mainEnabled, defaultOpts.scope.mainEnabled)) scopes = scopes.concat([
            mainScopeColor(
                `[${
                this.name
                ||
                require('path').basename(file.getFileName())
                }]`
            )
        ]);
        if (optOrDef(this.opts.scope.subEnabled, defaultOpts.scope.subEnabled) && child.name) scopes = scopes.concat([
            subScopeColor(
                `[${child.name}]`
            )
        ]);
        //End of Scope Generation

        //Icon Generation
        let iconStyle = generateStyle(this.opts.fail.icon.style || defaultOpts.fail.icon.style);
        let icon;
        if (this.opts.fail.icon.text) icon = this.opts.fail.icon.text + ' ';
        else icon = defaultOpts.fail.icon.text + ' ';

        if (icon == ' ') icon = '';

        icon = iconStyle(icon);
        //End of Icon Generation

        //Type Name Generation
        let nameStyle = generateStyle(this.opts.fail.style || defaultOpts.fail.style);
        let name = this.opts.fail.text || defaultOpts.fail.text;

        name = nameStyle(name);
        //End of Name Generation

        //Description Generation
        let descStyle = generateStyle(this.opts.fail.testDescStyle || defaultOpts.fail.testDescStyle);
        let description = child.description || `Line ${file.getLineNumber()}`;

        description = descStyle(description);
        //End of Description Generation

        //Detail Generation
        let detailStyle = generateStyle(this.opts.fail.detailStyle || defaultOpts.fail.detailStyle);
        let detailText = detailStyle(detail);
        //End of Detail Generation

        console.log(
            `${scopes.join(' ')} ${icon}${name} ${description} ${detailText}`
        );
    }
}

module.exports = {
    Test: Test,
    themes: {
        default: require('./themes/default.json'),
        halloween: require('./themes/holidays/halloween.json')
    }
};